/* $Id: video_entropyd.c,v 0.7 2012-05-09 11:11:28 folkert Exp $
 * $Log: not supported by cvs2svn $
 * Revision 0.6  2003/01/27 18:48:21  folkert
 * initial RCS check-in
 *
 */

#undef NO_CRONTAB

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include <syslog.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <sys/mman.h>
#include <linux/videodev2.h>
#include <linux/types.h>
#include <linux/random.h>

#include "error.h"
#include "RNGTEST.h"
#include "io.h"
#include "val.h"

#define DEFAULT_POOLSIZE_FN                     "/proc/sys/kernel/random/poolsize"
#define RNGTEST_PENALTY                         (20000 / 8) /* how many bytes to skip when the rng-test fails */

unsigned char *io_buffer;
unsigned int io_buffer_len;

int add_to_kernel_entropyspool(int handle, unsigned char *buffer, int nbytes)
{
        double nbits;
	int before;
        struct rand_pool_info *output;
        output = (struct rand_pool_info *)malloc(sizeof(struct rand_pool_info) + nbytes);
        if (!output)
        {
                error_exit("malloc failure in add_to_kernel_entropyspool");
                return -1;
        }

        // calculate number of bits in the block of
        // data. put in structure
        nbits = calc_nbits_in_data((unsigned char *)buffer, nbytes);
        if (nbits >= 1.0)
        {
		int after;

                output -> entropy_count = (int)nbits;
                output -> buf_size      = nbytes;
                memcpy(output -> buf, buffer, nbytes);

		printf("Adding %d bits of entropy-data to kernel entropy-spool.\n", (int)nbits);

		/* Get number of bits in KRNG before credit */
		if (ioctl(handle, RNDGETENTCNT, &before) == -1)
			error_exit("Coundn't query entropy to kernel");

                if (ioctl(handle, RNDADDENTROPY, output) == -1)
			error_exit("Coundn't add entropy-data to kernel");

		/* Get number of bits in KRNG after credit */
		if (ioctl(handle, RNDGETENTCNT, &after) == -1)
			error_exit("Coundn't query entropy to kernel");

		syslog(LOG_INFO, "Entropy credit of %i bits made (%i bits before, %i bits after)", output->entropy_count, before, after);
		printf("Entropy credit of %i bits made (%i bits before, %i bits after)\n", output->entropy_count, before, after);
        }
	else
		printf("Not enough entropy in data (%f bits)\n", nbits);

       	free(output);

	return (int)nbits;
}

long gen_rand(void)
{
	int fd;
	long dummy;

	fd = open("/dev/random", O_RDONLY);
	if (fd == -1)
	{
		fprintf(stderr, "Could not open /dev/random! (%d)\n", errno);
		exit(1);
	}

	(void)READ(fd, (char *)&dummy, sizeof(dummy));

	close(fd);

	return dummy;
}

char verify_is_capture_device(int fd)
{
	char rc = 1;
	struct v4l2_capability cap;
	memset(&cap, 0x00, sizeof(cap));

	if (ioctl(fd, VIDIOC_QUERYCAP, &cap) == -1)
	{
		fprintf(stderr, "Cannot VIDIOC_QUERYCAP %s\n", strerror(errno));
		rc = 0;
	}
	else
	{
		printf("%s / %s / %s\n", cap.driver, cap.card, cap.bus_info);

		if ((cap.capabilities & V4L2_CAP_VIDEO_CAPTURE) == 0)
		{
			fprintf(stderr, "Device can NOT do capture!\n");
			rc = 0;
		}
	}

	return rc;
}

int open_v4l2_dev(char *dev)
{
	enum v4l2_buf_type buf_type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	struct v4l2_requestbuffers req;
	struct v4l2_buffer buf;
	struct v4l2_format fmt;
	char format[5];
	struct v4l2_frmivalenum fr;
	int fd = open(dev, O_RDWR);
	if (fd == -1)
	{
		fprintf(stderr, "Cannot open %s\n", dev);
		exit(1);
	}

	if (!verify_is_capture_device(fd))
	{
		fprintf(stderr, "Not a capture device\n");
		exit(1);

	}

	memset(&req, 0x00, sizeof(req));
	req.count  = 1;
	req.type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	req.memory = V4L2_MEMORY_MMAP;

	memset(&buf, 0x00, sizeof(buf));
	buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buf.memory      = V4L2_MEMORY_MMAP;
	buf.index       = 0;

	memset(&fmt, 0x00, sizeof(fmt));
	fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUV420;
	if (ioctl(fd, VIDIOC_G_FMT, &fmt) == -1)
	{
		fprintf(stderr, "ioctl(VIDIOC_G_FMT) failed %s\n", strerror(errno));
		exit(1);
	}
	memcpy(format, &fmt.fmt.pix.pixelformat, 4);
	format[4]=0x00;
	printf(" %dx%d: %d/%s\n", fmt.fmt.pix.width, fmt.fmt.pix.height, fmt.type, format);

	memset(&fr, 0x00, sizeof(fr));
	fr.index = V4L2_FRMIVAL_TYPE_DISCRETE;
	fr.pixel_format = fmt.fmt.pix.pixelformat;
	fr.width = fmt.fmt.pix.width;
	fr.height = fmt.fmt.pix.height;

	fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if (ioctl(fd, VIDIOC_S_FMT, &fmt) == -1)
	{
		close(fd);
		fd = -1;
		fprintf(stderr, "ioctl(VIDIOC_S_FMT) failed %s\n", strerror(errno));
	}
	else if (ioctl(fd, VIDIOC_REQBUFS, &req) == -1)
	{
		close(fd);
		fd = -1;
		fprintf(stderr, "ioctl(VIDIOC_REQBUFS) failed %s\n", strerror(errno));
	}
	else if (ioctl(fd, VIDIOC_QUERYBUF, &buf) == -1)
	{
		close(fd);
		fd = -1;
		fprintf(stderr, "ioctl(VIDIOC_QUERYBUF) failed %s\n", strerror(errno));
	}
	else if (ioctl(fd, VIDIOC_QBUF, &buf) == -1)
	{
		close(fd);
		fd = -1;
		fprintf(stderr, "ioctl(VIDIOC_QBUF) failed %s\n", strerror(errno));
	}
	else if (ioctl(fd, VIDIOC_STREAMON, &buf_type) == -1)
	{
		close(fd);
		fd = -1;
		fprintf(stderr, "ioctl(VIDIOC_STREAMON) failed %s\n", strerror(errno));
	}
	else
	{
		io_buffer_len = buf.length;
		io_buffer = (unsigned char *)mmap(NULL, buf.length, PROT_READ | PROT_WRITE, MAP_SHARED, fd, buf.m.offset);
	}

	if (fd == -1)
		exit(1);

	return fd;
}

void take_picture(int fd)
{
	struct v4l2_buffer buf;
	do
	{
		memset(&buf, 0x00, sizeof(buf));
		buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory = V4L2_MEMORY_MMAP;
	}
	while(ioctl(fd, VIDIOC_DQBUF, &buf) == -1);

	if (ioctl(fd, VIDIOC_QBUF, &buf) == -1)
		fprintf(stderr, "VIDIOC_QBUF failed %s\n", strerror(errno));
}

void help(void)
{
	printf("-i x   device to use\n");
	printf("-s     ignore bad random\n");
	printf("-S x   number of seconds to sleep between each check\n");
	printf("-f     do not fork\n");
	printf("-V     show version and exit\n");
	printf("-h     show help and exit\n");
}

void version(void)
{
	printf("video_entropyd v" VERSION ", by folkert@vanheusden.com\n");
}

int main(int argc, char *argv[])
{
	unsigned char *prev, *unbiased;
	int nunbiased;
        int max_bits, todo, before, random_fd;
        FILE *poolsize_fh;
	unsigned int loop;
	int c;
	char *device = "/dev/video0";
	char ignore_bad_random = 0;
	unsigned int device_settle = 25;
	int sleep_delay = 3000;
	char fork = 1;
	int vdev_fd = -1;

	while((c = getopt(argc, argv, "fi:sS:Vh")) != -1)
	{
		switch(c)
		{
			case 'f':
				fork = 0;
				break;

			case 'i':
				device = optarg;
				break;

			case 's':
				ignore_bad_random = 1;
				break;

			case 'S':
				sleep_delay = atoi(optarg);
				break;

			case 'V':
				version();
				return 0;

			case 'h':
				help();
				return 0;

			default:
				help();
				return 1;
		}
	}

	if (ignore_bad_random)
		printf("Ignoring bad random data\n");

        /* find out poolsize */
        poolsize_fh = fopen(DEFAULT_POOLSIZE_FN, "rb");
        if (!poolsize_fh)
        {
                syslog(LOG_CRIT, "Couldn't open poolsize file: %m");
                exit(1);
        }
        fscanf(poolsize_fh, "%d", &max_bits);
        fclose(poolsize_fh);
        max_bits = (max_bits << 3); /* bytes -> bits */

	/* open random device and write entropy-data to it */
	random_fd = open("/dev/random", O_WRONLY);
	if (random_fd == -1)
	{
		fprintf(stderr, "Could not open /dev/random! (%d)\n", errno);
		exit(1);
	}

	/* open device */
	vdev_fd = open_v4l2_dev(device);

	/* allocate memory for difference image & previous image */
	prev = (unsigned char *)malloc(io_buffer_len);
	unbiased = (unsigned char *)malloc(io_buffer_len);

	if (!prev || !unbiased)
	{
		printf("Could not allocate enough memory! (%d bytes)\n", io_buffer_len*2);
		return 1;
	}

	/* let device settle */
	for(loop=0; loop<device_settle; loop++)
	{
		printf("waiting for device to settle... %d \r", device_settle - loop);
		fflush(NULL);

		take_picture(vdev_fd);
		usleep(40000);
	}
	printf("\n");

	/* take first picture */
	take_picture(vdev_fd);
	memcpy(prev, io_buffer, io_buffer_len);

	RNGTEST_init();

	printf("Go!\n");

	if (fork == 1 && daemon(-1, -1) == -1)
	{
		syslog(LOG_ALERT, "Could not become daemon");
		exit(1);
	}

	/* main loop */
	/* no for-loop in case you want to be able to take pictures
	 * while using this program
	 */
	for(;;)
	{
		/* find out how many bits to add */
		if (ioctl(random_fd, RNDGETENTCNT, &before) == -1)
		{
			syslog(LOG_CRIT, "Coundn't query entropy to kernel");
			exit(1);
		}
		todo = max_bits - before;

		printf("need to get %d bits\n", todo);

		while(todo > 0)
		{
			struct timespec req, rem;
			unsigned char byte=0, nbits=0;
			static int error_state = 0;

			/* sleep a while */
			req.tv_sec = (gen_rand() % 5) + 1;
			req.tv_nsec = gen_rand() % 999999999;
			for(;;)
			{
				if (nanosleep(&req, &rem) == 0)
					break;
				if (errno != EINTR)
					break;

				req.tv_sec  = rem.tv_sec;
				req.tv_nsec = rem.tv_nsec;
			}

			/* take a picture */
			printf("smile!\n");
			take_picture(vdev_fd);

			/* unbiase */
			nunbiased=0;
			for(loop=0; loop<io_buffer_len; loop+=2)
			{
				/* calculate difference between the images */
				int diff1 = abs(io_buffer[loop] - prev[loop]);
				int diff2 = abs(io_buffer[loop+1] - prev[loop+1]);

				/* if the 2 difference are not correlated, add bit */
				if ((diff1 & 1) != (diff2 & 1))
				{
					byte <<= 1;

					if (diff1 & 1)
						byte |= 1;

					nbits++;

					if (nbits == 8)
					{
						if (error_state == 0)
						{
							unbiased[nunbiased++] = byte;
						}
						nbits = 0;

						if (!ignore_bad_random)
						{
							RNGTEST_add(byte);
							if (RNGTEST() == -1)
							{
								if (error_state == 0)
								{
									printf("test of random data failed, skipping %d bytes before re-using data-stream (%d bytes in flush)\n", RNGTEST_PENALTY, error_state);
									syslog(LOG_CRIT, "test of random data failed, skipping %d bytes before re-using data-stream (%d bytes in flush)", RNGTEST_PENALTY, error_state);
								}
								error_state = RNGTEST_PENALTY;
								nunbiased = 0;
							}
							else
							{
								if (error_state > 0)
								{
									error_state--;

									if (error_state == 0)
									{
										printf("Restarting fetching of entropy data\n");
										syslog(LOG_INFO, "Restarting fetching of entropy data");
									}
								}
							}
						}
					}
				}
			}

			if (nunbiased > 0)
			{
				printf("Submitting %d bytes", nunbiased);
				todo -= add_to_kernel_entropyspool(random_fd, unbiased, nunbiased) / 8;
				if (todo > 0)
					printf(", %d bytes left.\n", todo);
				else
					printf(", done.\n");
			}
			else if (error_state > 0)
			{
				printf("%d bytes left to ignore.\n", error_state);
			}

			/* remember picture */
			memcpy(prev, io_buffer, io_buffer_len);
		}

		printf("Sleeping...\n");

		sleep(sleep_delay);
	}

	close(random_fd);
	close(vdev_fd);

	return 0;
}
